@@ -297,25 +297,6 @@ find_keycode (GkbdKeyboardDrawing * drawing, gchar * key_name)
 }
 
 
-static void
-fit_width (GkbdKeyboardDrawing * drawing, gint width)
-{
-	PangoRectangle logical_rect;
-	gint old_size;
-
-	pango_layout_get_extents (drawing->layout, NULL, &logical_rect);
-
-	if (logical_rect.width > 0 && logical_rect.width > width) {
-		old_size =
-		    pango_font_description_get_size (drawing->font_desc);
-		pango_font_description_set_size (drawing->font_desc,
-						 old_size * width /
-						 logical_rect.width);
-		pango_layout_set_font_description (drawing->layout,
-						   drawing->font_desc);
-	}
-}
-
 static void
 set_key_label_in_layout (GkbdKeyboardDrawing * drawing,
 			 PangoLayout * layout, guint keyval)
@@ -485,7 +466,7 @@ set_key_label_in_layout (GkbdKeyboardDrawing * drawing,
 
 
 static void
-draw_layout (GkbdKeyboardDrawing * drawing,
+draw_pango_layout (GkbdKeyboardDrawing * drawing,
 	     gint angle, gint x, gint y, PangoLayout * layout)
 {
 	GtkStateType state = GTK_WIDGET_STATE (GTK_WIDGET (drawing));
@@ -521,16 +502,10 @@ draw_layout (GkbdKeyboardDrawing * drawing,
 
 			for (j = 0; j < run->glyphs->num_glyphs; j++) {
 				PangoGlyphGeometry *geometry;
-				gint xx, yy;
 
 				geometry =
 				    &run->glyphs->glyphs[j].geometry;
 
-				rotate_coordinate (0, 0, x_off, y_off,
-						   angle, &xx, &yy);
-				geometry->x_offset -= x_off - xx;
-				geometry->y_offset -= y_off - yy;
-
 				x_off += geometry->width;
 			}
 		}
@@ -554,7 +529,6 @@ draw_key_label_helper (GkbdKeyboardDrawing * drawing,
 		       gint x,
 		       gint y, gint width, gint height, gint padding)
 {
-	gint old_size;
 	gint label_x, label_y, label_max_width, ycell;
 
 	if (keysym == 0)
@@ -603,19 +577,10 @@ draw_key_label_helper (GkbdKeyboardDrawing * drawing,
 		return;
 	}
 	set_key_label_in_layout (drawing, drawing->layout, keysym);
-
-	old_size = pango_font_description_get_size (drawing->font_desc);
-	fit_width (drawing, label_max_width);
-
-	draw_layout (drawing, angle, label_x, label_y, drawing->layout);
-
-	if (pango_font_description_get_size (drawing->font_desc) !=
-	    old_size) {
-		pango_font_description_set_size (drawing->font_desc,
-						 old_size);
-		pango_layout_set_font_description (drawing->layout,
-						   drawing->font_desc);
-	}
+	pango_layout_set_width (drawing->layout, label_max_width);
+	label_y -= (pango_layout_get_line_count (drawing->layout) - 1) *
+		(pango_font_description_get_size (drawing->font_desc) / PANGO_SCALE);
+	draw_pango_layout (drawing, angle, label_x, label_y, drawing->layout);
 }
 
 static void
@@ -654,6 +619,13 @@ draw_key_label (GkbdKeyboardDrawing * drawing,
 		    || l >= XkbKeyGroupWidth (drawing->xkb, keycode, g))
 			continue;
 
+		/* Skip "exotic" levels like the "Ctrl" level in PC_SYSREQ */
+		if (l > 0) {
+			guint mods = XkbKeyKeyType (drawing->xkb, keycode, g)->mods.mask;
+			if ((mods & (ShiftMask | drawing->l3mod)) == 0)
+				continue;
+		}
+
 		if (drawing->track_modifiers) {
 			uint mods_rtrn;
 			KeySym keysym;
@@ -688,7 +660,7 @@ draw_key (GkbdKeyboardDrawing * drawing, GkbdKeyboardDrawingKey * key)
 {
 	XkbShapeRec *shape;
 	GdkColor *color;
-	gint i;
+	/* gint i; */
 
 	if (!drawing->xkb)
 		return;
@@ -713,9 +685,22 @@ draw_key (GkbdKeyboardDrawing * drawing, GkbdKeyboardDrawingKey * key)
 		shape->num_outlines);
 #endif
 
-	for (i = 0; i < 1 /* shape->num_outlines */ ; i++)
-		draw_outline (drawing, shape->outlines + i, color,
-			      key->angle, key->origin_x, key->origin_y);
+        /* draw the primary outline */
+        draw_outline (drawing, shape->primary ? shape->primary : shape->outlines, 
+		      color, key->angle, key->origin_x, key->origin_y);
+#if 0
+        /* don't draw other outlines for now, since
+         * the text placement does not take them into account 
+         */
+        for (i = 0; i < shape->num_outlines; i++)
+          {
+            if (shape->outlines + i == shape->approx ||
+	        shape->outlines + i == shape->primary)
+	      continue;
+            draw_outline (drawing, shape->outlines + i, NULL, 
+		          key->angle, key->origin_x, key->origin_y);
+	  }
+#endif
 
 	draw_key_label (drawing, key->keycode, key->angle, key->origin_x,
 			key->origin_y, shape->bounds.x2, shape->bounds.y2);
@@ -815,7 +800,7 @@ draw_text_doodad (GkbdKeyboardDrawing * drawing,
 				 doodad->origin_y + text_doodad->top);
 
 	pango_layout_set_text (drawing->layout, text_doodad->text, -1);
-	draw_layout (drawing, doodad->angle, x, y, drawing->layout);
+	draw_pango_layout (drawing, doodad->angle, x, y, drawing->layout);
 }
 
 static void
@@ -858,11 +843,21 @@ draw_shape_doodad (GkbdKeyboardDrawing * drawing,
 	shape = drawing->xkb->geom->shapes + shape_doodad->shape_ndx;
 	color = drawing->colors + shape_doodad->color_ndx;
 
-	for (i = 0; i < shape->num_outlines; i++)
-		draw_outline (drawing, shape->outlines + i, color,
-			      doodad->angle,
-			      doodad->origin_x + shape_doodad->left,
-			      doodad->origin_y + shape_doodad->top);
+        /* draw the primary outline filled */
+        draw_outline (drawing, shape->primary ? shape->primary : shape->outlines, 
+ 		      color, doodad->angle,
+                      doodad->origin_x + shape_doodad->left,
+		      doodad->origin_y + shape_doodad->top);
+
+        /* stroke the other outlines */
+	for (i = 0; i < shape->num_outlines; i++) {
+                if (shape->outlines + i == shape->approx ||
+	            shape->outlines + i == shape->primary) 
+	              continue;
+                draw_outline (drawing, shape->outlines + i, NULL, doodad->angle,
+		              doodad->origin_x + shape_doodad->left,
+		              doodad->origin_y + shape_doodad->top);
+        }
 }
 
 static void
@@ -949,6 +944,7 @@ alloc_pango_layout (GkbdKeyboardDrawing * drawing)
 	PangoContext *context =
 	    gtk_widget_get_pango_context (GTK_WIDGET (drawing));
 	drawing->layout = pango_layout_new (context);
+	pango_layout_set_ellipsize (drawing->layout, PANGO_ELLIPSIZE_END);
 }
 
 static void
@@ -968,7 +964,7 @@ expose_event (GtkWidget * widget,
 		return FALSE;
 
 	if (drawing->pixmap == NULL)
-		draw_keyboard (drawing);
+		return FALSE;
 
 	gdk_draw_drawable (widget->window,
 			   widget->style->fg_gc[state],
@@ -988,6 +984,17 @@ expose_event (GtkWidget * widget,
 	return FALSE;
 }
 
+static gboolean
+idle_redraw (gpointer user_data)
+{
+	GkbdKeyboardDrawing *drawing = user_data;
+
+	drawing->idle_redraw = 0;
+	draw_keyboard (drawing);
+	gtk_widget_queue_draw (GTK_WIDGET (drawing));
+	return FALSE;
+}
+
 static void
 size_allocate (GtkWidget * widget,
 	       GtkAllocation * allocation, GkbdKeyboardDrawing * drawing)
@@ -1024,6 +1031,9 @@ size_allocate (GtkWidget * widget,
 				  drawing->scale_denominator);
 	pango_layout_set_font_description (drawing->layout,
 					   drawing->font_desc);
+
+	if (!drawing->idle_redraw)
+		drawing->idle_redraw = g_idle_add (idle_redraw, drawing);
 }
 
 static gint
@@ -1488,6 +1498,10 @@ destroy (GkbdKeyboardDrawing * drawing)
 		g_source_remove (drawing->timeout);
 		drawing->timeout = 0;
 	}
+	if (drawing->idle_redraw > 0) {
+		g_source_remove (drawing->idle_redraw);
+		drawing->idle_redraw = 0;
+	}
 
 	g_object_unref (drawing->pixmap);
 }
@@ -1558,6 +1572,8 @@ gkbd_keyboard_drawing_init (GkbdKeyboardDrawing * drawing)
 	}
 
 	XkbGetNames (drawing->display, XkbAllNamesMask, drawing->xkb);
+	drawing->l3mod = XkbKeysymToModifiers (drawing->display,
+					       GDK_ISO_Level3_Shift);
 
 	drawing->xkbOnDisplay = TRUE;
 
